
/*
	TrayMin - PC Magazine Windows tray minimization utility

	Copyright (c) 2000 Ziff-Davis Publishing Company.  All rights reserved.
	First published in PC Magazine, US Edition  .

	Written by Steven E. Sipe


	This file implements TrayMin's main menu handler and configuration
	routine.
*/

#include <windows.h>

#include <string.h>
#include <stdio.h>

#include "trayminz.h"
#include "resource.h"
#include "tmzhook.h"
#include "tray.h"

// Prototypes for local routines
LRESULT CALLBACK About(HWND, UINT, WPARAM, LPARAM);

// Private data declarations
static HINSTANCE g_hInstance;
static HINSTANCE g_hLib;
static TRAYDATA g_aTD[MAX_TRAYAPPS];
static HWND g_hWndMain;
static HWND g_hWndPopup;
static BOOL g_bInMenu = FALSE;

// Returns the location of TrayMin's help file (same location as .exe)
LPCTSTR GetHelpFileName()
{
	static char szDrive[MAX_PATH+1], szPath[MAX_PATH+1];
 	char szDummy[MAX_PATH+1], szAppName[MAX_PATH+1];

	// Get the .exe's full path
	GetModuleFileName(NULL,szAppName,sizeof(szAppName));
	_splitpath(szAppName,szDrive,szPath,szDummy,szDummy);

	// Save the driver and path and append the help file name
	strcat(szDrive,szPath);
	strcat(szDrive,"traymin.hlp");

	return(szDrive);
}

// Returns TRUE if TrayMin is already running
BOOL AlreadyRunning()
{
   BOOL bFound = FALSE;

   // Try to create a mutex with the app's name
   HANDLE hMutexOneInstance = CreateMutex(NULL,TRUE,"TrayMin");

   // Already there...means that we are already running an instance
   if(GetLastError() == ERROR_ALREADY_EXISTS)
      bFound = TRUE;

   // Release the mutex
   if(hMutexOneInstance)
      ReleaseMutex(hMutexOneInstance);

   return(bFound);
}

// Hash the module's name 
unsigned long HashName(LPCTSTR lpszName)
{
	unsigned long ulHash = 0;

	while(*lpszName)
		ulHash = (ulHash << 5)+ulHash+*lpszName++;

	return ulHash;
}

// Returns the icon and title for the given window
HICON GetTrayLabel(HWND hWnd, LPCTSTR lpszFileName, LPTSTR lpszTitle, int nMaxLen)
{
	HICON hIcon, hIconLarge = NULL, hIconSmall = NULL;

	// Get the app's name
	GetWindowText(hWnd,lpszTitle,nMaxLen);

	// Get the app's icon, both large and small
	ExtractIconEx(lpszFileName,0,&hIconLarge,&hIconSmall,1);

	if(hIconSmall)
		hIcon = hIconSmall;
	else if(hIconLarge)
		hIcon = hIconLarge;
	else hIcon = LoadIcon(NULL,IDI_WINLOGO);

	return hIcon;
}

// Search the list of tray icon IDs and see if the given ID is already
// in use
int FindID(unsigned long ulID)
{
	int i;

	// Iterate through slots
	for(i = 0; i < MAX_TRAYAPPS; i++)
	{
		// Found a slot?
		if(g_aTD[i].hWnd && g_aTD[i].ulHash == ulID)
			return(i);
	}

	return(-1);
}

// Performs a ShowWindow() call without locking up on frozen applications.
// Basically, the WM_SHOWWINDOW message generated for each app by ShowWindow() 
// locks the sending app under Windows if the receiving app is dead
BOOL SafeShowWindow(HWND hWnd, int nShowCmd)
{
   // Determine if the app is still alive or hung
   DWORD dwResult;
   if(SendMessageTimeout(hWnd,WM_NULL,0,0,SMTO_ABORTIFHUNG,2000,&dwResult))
   {
      // Seemed to work okay --- show the window
      if(dwResult == 0)
         ShowWindow(hWnd,nShowCmd);
      return(TRUE);
   }
   else return(FALSE);
}

// Called for each window in a thread -- looks for a Delphi style
// main window so we can handle this special case.  Delphi 
// apps actually have *two* main windows.  The one that 
// controls the start bar is of the TApplication class
BOOL CALLBACK EnumThreadWndProc(HWND hWnd, LPARAM lParam)
{
	char szClassName[500];

	// Only need to check visible windows
	if(IsWindowVisible(hWnd))
	{
		// Get the class name
		GetClassName(hWnd,szClassName,sizeof(szClassName));

		// Is it a Delphi main window?
		if(!strcmp(szClassName,"TApplication"))
		{
			HWND *phWndDelphi = (HWND *) lParam;

			// Yep...save it
			*phWndDelphi = hWnd;

			// This stops the enumeration
			return(FALSE);
		}
	}

	// Continue the enumeration
	return(TRUE);
}

// Handles the special case of Delphi application windows
HWND FindDelphiWindow(HWND hWnd)
{
	DWORD dwTID = GetWindowThreadProcessId(hWnd,NULL);
	HWND hWndDelphi = NULL;

	// Search through each window created by this thread
	EnumThreadWindows(dwTID,EnumThreadWndProc,(LPARAM) &hWndDelphi);

	// Return the resulting handle
	return(hWndDelphi);
}

// Adds an app to the hashed ID collection
BOOL AddSlot(HWND hWnd, LPCTSTR lpszFileName)
{
	int i, nIndex = 0;
	BOOL bRet = FALSE;
	HWND hWndTemp;

	// Iterate through slots
	for(i = 0; !bRet && i < MAX_TRAYAPPS; i++)
	{
		// Found a slot?
		if(!g_aTD[i].hWnd)
		{
			HICON hIcon;
			char szTitle[500], szIndex[10];
			unsigned long ulHash = HashName(lpszFileName);
			int nTries = 50;

			// Is this number in use?  If so, increment the
			// number by one.  (NOTE: this could cause collisions
			// too, but it's worth the risk since it allows 
			// TrayManager to work well with TrayMin)
			while(FindID(ulHash) >= 0 && nTries--)
			{
				nIndex++;
				ulHash++;
			}

			// Determine if there's an "extra" Delphi window for this
			// app
			hWndTemp = FindDelphiWindow(hWnd);

			// If there's a hidden Delphi window then the "real" TApplication
			// window is actually the main window and the secondary window
			// becomes the value in hWnd.  Otherwise, we're dealing with a
			// normal application that has a main window (hWnd) and the secondary
			// window (hWndDelphi) is set to NULL
			g_aTD[i].hWnd = hWndTemp?hWndTemp:hWnd;
			g_aTD[i].hWndDelphi = hWndTemp?hWnd:NULL;

			// Save the filename a calculated has code
			strcpy(g_aTD[i].szFileName,lpszFileName);
			g_aTD[i].ulHash = ulHash;

			// Hide the main window and Delphi window if necessary
			SafeShowWindow(g_aTD[i].hWnd,SW_HIDE);
			if(g_aTD[i].hWndDelphi)
				SafeShowWindow(g_aTD[i].hWndDelphi,SW_HIDE);

			// Get the tray icon's default information
			hIcon = GetTrayLabel(g_aTD[i].hWnd,lpszFileName,szTitle,sizeof(szTitle));

			// Determine if we're dealing with a multi-instance app.  If so, 
			// append the instance count to its tooltip title
			if(nIndex > 0)
			{
				sprintf(szIndex," #%d",nIndex+1);
				strcat(szTitle,szIndex);
			}

			// Finally, add the new tray icon
			Tray_Add(g_hWndMain,UM_TRAYMSG,g_aTD[i].ulHash,szTitle,hIcon);

			bRet = TRUE;
		}
	}

	return(bRet);
}

// Removes an ID from the hashed ID collection
BOOL RemoveSlot(ULONG ulID)
{
	int i;
	BOOL bRet = FALSE;

	// Iterate through slots
	for(i = 0; !bRet && i < MAX_TRAYAPPS; i++)
	{
		// Found a slot?
		if(g_aTD[i].hWnd && (g_aTD[i].ulHash == ulID || !ulID))
		{
			// Restore the associated application
			ShowWindow(g_aTD[i].hWnd,SW_SHOW);
			if(g_aTD[i].hWndDelphi)
				ShowWindow(g_aTD[i].hWndDelphi,SW_SHOW);

			if(ulID)
				SetForegroundWindow(g_aTD[i].hWnd);
			Tray_Delete(g_hWndMain,g_aTD[i].ulHash);

			// Reset the slot
			g_aTD[i].hWnd = 0;
			g_aTD[i].hWndDelphi = 0;

			// Get out now, if we're only looking for one ID
			if(ulID)
				bRet = TRUE;
		}
	}

	return(bRet);
}

// Display a "pseudo" system menu for the tray icon
void ShowSystemMenu(unsigned long ulID)
{
	POINT ptMouse;
	HMENU hMenu;
	UINT uCmd ;
	HWND hWnd;
	int nIndex = FindID(ulID);
	DWORD dwStyle;
	int i;

	if(nIndex < 0)
		return;

	// Get the cursor's position
	GetCursorPos(&ptMouse);

	// Get the window handle
	hWnd = g_aTD[nIndex].hWnd;

	// Get the window's style
	dwStyle = GetWindowLong(hWnd,GWL_STYLE);

	// Get its system menu handle
	SetForegroundWindow(g_hWndMain);
	hMenu = GetSystemMenu(hWnd,FALSE);

	// Enable the Restore menu item and make it the default
	EnableMenuItem(hMenu,SC_RESTORE,MF_BYCOMMAND|MF_ENABLED);
	SetMenuDefaultItem(hMenu,SC_RESTORE,FALSE);

	// Set the proper state for the maximize menu item
	EnableMenuItem(hMenu,SC_MAXIMIZE,MF_BYCOMMAND|(dwStyle & WS_MAXIMIZEBOX?MF_ENABLED:MF_DISABLED|MF_GRAYED));

	// Disable the Size and Move menu items
	EnableMenuItem(hMenu,SC_MOVE,MF_BYCOMMAND|MF_DISABLED|MF_GRAYED);
	EnableMenuItem(hMenu,SC_SIZE,MF_BYCOMMAND|MF_DISABLED|MF_GRAYED);

	// Display the system menu
	uCmd = TrackPopupMenuEx(hMenu,TPM_RETURNCMD|TPM_LEFTBUTTON,
							ptMouse.x,ptMouse.y,g_hWndMain,NULL);

	// Was a valid selection made?
	if(uCmd)
	{
		// Yes...remove the tray icon, restore the window and
		// execute the command
		for(i = 0; i < MAX_TRAYAPPS; i++)
		{
			// Found a slot?
			if(g_aTD[i].hWnd && (g_aTD[i].ulHash == ulID))
			{
				// Reset the slot
				g_aTD[i].hWnd = 0;
				break;
			}
		}

		// Delete the tray icon and show the window again
		Tray_Delete(g_hWndMain,ulID);
		ShowWindow(hWnd,SW_SHOW);
		if(g_aTD[i].hWndDelphi)
			ShowWindow(g_aTD[i].hWndDelphi,SW_SHOW);
		UpdateWindow(hWnd);

		// Send the system menu command to the application
		SendMessage(hWnd,WM_SYSCOMMAND,uCmd,0);
		if(g_aTD[i].hWndDelphi)
			SendMessage(g_aTD[i].hWndDelphi,WM_SYSCOMMAND,uCmd,0);
	}

	// Get rid of the system menu
	PostMessage(g_hWndMain,WM_NULL,0,0);
}

// Displays the tray app's system menu.  This is done when the 
// user right clicks on the app's tray icon
void ShowContextMenu()
{
	HMENU hMenuMain, hMenuContext;
	POINT ptMouse;

	// Already in the menu?
	if(g_bInMenu) 
	{
		// Yep...complain and get out
		MessageBeep(0);
		BringWindowToTop(g_hWndMain);
		SetForegroundWindow(g_hWndMain);
		return;
	}

	// Load the context menu
	hMenuMain = LoadMenu(g_hInstance,MAKEINTRESOURCE(IDR_TRAYMENU));
	hMenuContext = GetSubMenu(hMenuMain,0);

	// Make sure our window is activated
	SetForegroundWindow(g_hWndMain);

	// Display the context menu near the mouse cursor
	GetCursorPos(&ptMouse);
	TrackPopupMenu(hMenuContext,
               TPM_BOTTOMALIGN|TPM_LEFTBUTTON|TPM_RIGHTBUTTON,
               ptMouse.x,ptMouse.y,0,g_hWndMain,NULL);

	// We no longer need the menu so destroy it
	DestroyMenu(hMenuMain);

	// Release the context menu
	PostMessage(g_hWndMain,WM_NULL,0,0);
}

// TrayMin's main window procedure
long FAR PASCAL MainWndProc(HWND hWnd, unsigned message, WPARAM wParam, 
										LPARAM lParam)
{
	static HWND hWndMenu = NULL;

	switch(message)
	{
		case WM_CREATE:
			{
				// Save TrayMin's window handle
				g_hWndMain = hWnd;
	
				// Set the notify window for the hook
				SetNotifyWindow(hWnd);

				// Set the hook
				SetHook();

				// Add TM's tray icon
				if(__argc < 2 || (stricmp(__argv[1],"-i") && stricmp(__argv[1],"/i")))
					Tray_Add(hWnd,UM_TRAYMSG,1,"TrayMin",
							LoadIcon(GetModuleHandle(NULL),
							MAKEINTRESOURCE(IDI_TRAYMINZ)));
			}	
			break;

		// Retrieve data from hook DLL
		case WM_COPYDATA:
			{
				COPYDATASTRUCT *pCDS = (COPYDATASTRUCT *) lParam;

				// Pass along the tray message if we have one that's valid
				if(pCDS)
				{
					TRAYDATA *pTD = (TRAYDATA *) pCDS->lpData;

					// Add a new hashed ID slot
					if(pTD)
						AddSlot(pTD->hWnd,pTD->szFileName);
				}

				return(TRUE);
			}
			break;

		case WM_COMMAND:
			switch(LOWORD(wParam))
			{
				// Display the About box
				case IDM_ABOUT:
					g_bInMenu = TRUE;
					DialogBox(g_hInstance,(LPCTSTR)IDD_ABOUTBOX,hWnd, 
								(DLGPROC)About);
					g_bInMenu = FALSE;
					break;

				// Display the help window
				case IDM_HELP:
					WinHelp(hWnd,GetHelpFileName(),HELP_FINDER,0);
					break;

				// Exit
				case IDM_EXIT:
					PostMessage(hWnd,WM_CLOSE,0,0);
					break;
			}
			break;

		// Handle messages for icons in the Windows tray
		case UM_TRAYMSG:
			switch(lParam)
			{
				// Double-click invokes the Configure screen
				case WM_LBUTTONDBLCLK:
					break;

				// Active app
				case WM_LBUTTONUP:
					if(wParam)
					{
						RemoveSlot((unsigned long) wParam);
					}
					break;

				// Display app's system menu
				case WM_RBUTTONUP:
					// Is this TrayMin's icon??
					if(wParam == 1)	
						ShowContextMenu();
					else ShowSystemMenu(wParam);
					break;
			}
			break;

		// Main window is going away...exit TM now
		case WM_DESTROY:
			// Restore all of the minimized applications
			RemoveSlot(0);

			// Remove traymin's icon
			Tray_Delete(hWnd,1);

			// Terminate the message hook
			ResetHook();

			// Quit
			PostQuitMessage(0);
			break;

		default:
			return (DefWindowProc(hWnd, message, wParam, lParam));
	}

	return(0);
}

// Register TM's main application class
BOOL InitApplication()
{
	WNDCLASS  wc;

	wc.style = CS_DBLCLKS;                     
	wc.lpfnWndProc = MainWndProc;        
	wc.cbClsExtra = 0;                   
	wc.cbWndExtra = 0;                   
	wc.hInstance = g_hInstance;            
	wc.hbrBackground = (HBRUSH) (COLOR_WINDOW+1);
	wc.lpszMenuName =  NULL;      
	wc.hIcon = NULL;
	wc.hCursor = LoadCursor(NULL,IDC_ARROW);
	wc.lpszClassName = "TrayMin";         

	/* Register the window class and return success/failure code. */
	return (RegisterClass(&wc));
}

// Create this specific instance of TM
BOOL InitInstance(int nCmdShow)
{
	HWND hWnd;               
	DWORD dwStyle;

	dwStyle = WS_OVERLAPPED|WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX|WS_THICKFRAME;

	 /* Create a main window for this application instance.  */
	hWnd = CreateWindow(
			"TrayMin",                     
			"TrayMin",      
			dwStyle,
			CW_USEDEFAULT,                 
			CW_USEDEFAULT,                 
			500,       
			50,                           
			NULL,                          
			NULL,                          
			g_hInstance,                     
			NULL                           
		);

	if(!hWnd)
		return (FALSE);

	 /* Make the window visible; update its client area; and return "success" */
	UpdateWindow(hWnd);
  
	return (TRUE);             
}

// Mesage handler for about box.
LRESULT CALLBACK About(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
		case WM_INITDIALOG:
				g_hWndPopup = hDlg;
				return TRUE;

		case WM_COMMAND:
			if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL) 
			{
				EndDialog(hDlg, LOWORD(wParam));
				return TRUE;
			}
			break;
	}
    return FALSE;
}

// TM's main entry point
int WINAPI WinMain(HINSTANCE hInstance,
			HINSTANCE hPrevInstance,
			LPSTR lpCmdLine,
			int nCmdShow)
{
	MSG msg;                            

	g_hInstance = hInstance;

	// Load TM's hook DLL
	g_hLib = LoadLibrary("tmzhook.dll");

	// Did it load??
	if(g_hLib == NULL)
	{
		// Nope...whine about it and get out
		MessageBox(NULL,"Unable to open tmzhook.dll.","TrayMin",
								MB_ICONERROR|MB_OK);
		return(FALSE);
	}

	// Is TM already running?
	if(AlreadyRunning())
	{
		MessageBox(NULL,"TrayMin is already running.","TrayMin",
								MB_ICONWARNING|MB_OK);
		return(0);
	}

	// Clear the tray data structures
	memset(g_aTD,0,sizeof(g_aTD));

	// Initialize TM
	if(!InitApplication())    
		return(-1);               

	// Initialize TM's instance
	if(!InitInstance(SW_SHOWNORMAL))
		return(-1);

	// Handle TM messages
	while(GetMessage(&msg,0,0,0))                       
	{
		TranslateMessage(&msg);           
		DispatchMessage(&msg);            
	}                       

	FreeLibrary(g_hLib);
 
	return(0); 
}

